@@ -21,18 +21,42 @@
 
 """The Desktop Couch Records API."""
 
+import httplib2, urlparse, cgi, copy
+from time import time
+
+# please keep desktopcouch python 2.5 compatible for now
+
+# pylint can't deal with failing imports even when they're handled
+# pylint: disable-msg=F0401
+try:
+    # Python 2.5
+    import simplejson as json
+except ImportError:
+    # Python 2.6+
+    import json
+# pylint: enable-msg=F0401
+
+from oauth import oauth
+
 from couchdb import Server
-from couchdb.client import ResourceNotFound, ResourceConflict
+from couchdb.client import ResourceNotFound, ResourceConflict, uri as couchdburi
 from couchdb.design import ViewDefinition
 from record import Record
-import httplib2
-from oauth import oauth
-import urlparse
-import cgi
 
 #DEFAULT_DESIGN_DOCUMENT = "design"
 DEFAULT_DESIGN_DOCUMENT = None  # each view in its own eponymous design doc.
 
+class FieldsConflict(Exception):
+    """Raised in case of an unrecoverable couchdb conflict."""
+
+    #pylint: disable-msg=W0231
+    def __init__(self, conflicts):
+        self.conflicts = conflicts
+    #pylint: enable-msg=W0231
+
+    def __str__(self):
+        return "<CouchDB Conflict Error: %s>" % self.conflicts
+
 
 class NoSuchDatabase(Exception):
     "Exception for trying to use a non-existent database"
@@ -45,21 +69,22 @@ class NoSuchDatabase(Exception):
         return ("Database %s does not exist on this server. (Create it by "
                 "passing create=True)") % self.database
 
+
 class OAuthAuthentication(httplib2.Authentication):
     """An httplib2.Authentication subclass for OAuth"""
-    def __init__(self, oauth_data, host, request_uri, headers, response, 
+    def __init__(self, oauth_data, host, request_uri, headers, response,
         content, http, scheme):
         self.oauth_data = oauth_data
         self.scheme = scheme
-        httplib2.Authentication.__init__(self, None, host, request_uri, 
+        httplib2.Authentication.__init__(self, None, host, request_uri,
               headers, response, content, http)
 
     def request(self, method, request_uri, headers, content):
         """Modify the request headers to add the appropriate
         Authorization header."""
-        consumer = oauth.OAuthConsumer(self.oauth_data['consumer_key'], 
+        consumer = oauth.OAuthConsumer(self.oauth_data['consumer_key'],
            self.oauth_data['consumer_secret'])
-        access_token = oauth.OAuthToken(self.oauth_data['token'], 
+        access_token = oauth.OAuthToken(self.oauth_data['token'],
            self.oauth_data['token_secret'])
         sig_method = oauth.OAuthSignatureMethod_HMAC_SHA1
         full_http_url = "%s://%s%s" % (self.scheme, self.host, request_uri)
@@ -76,14 +101,15 @@ class OAuthAuthentication(httplib2.Authentication):
         req.sign_request(sig_method(), consumer, access_token)
         headers.update(httplib2._normalize_headers(req.to_header()))
 
+
 class OAuthCapableHttp(httplib2.Http):
-    """Subclass of httplib2.Http which specifically uses our OAuth 
+    """Subclass of httplib2.Http which specifically uses our OAuth
        Authentication subclass (because httplib2 doesn't know about it)"""
     def __init__(self, scheme="http", cache=None, timeout=None, proxy_info=None):
         self.__scheme = scheme
         super(OAuthCapableHttp, self).__init__(cache, timeout, proxy_info)
 
-    def add_oauth_tokens(self, consumer_key, consumer_secret, 
+    def add_oauth_tokens(self, consumer_key, consumer_secret,
                          token, token_secret):
         self.oauth_data = {
             "consumer_key": consumer_key,
@@ -97,7 +123,7 @@ class OAuthCapableHttp(httplib2.Http):
         """Since we know we're talking to desktopcouch, and we know that it
            requires OAuth, just return the OAuthAuthentication here rather
            than checking to see which supported auth method is required."""
-        yield OAuthAuthentication(self.oauth_data, host, request_uri, headers, 
+        yield OAuthAuthentication(self.oauth_data, host, request_uri, headers,
                 response, content, self, self.__scheme)
 
 def row_is_deleted(row):
@@ -125,6 +151,9 @@ class CouchDatabaseBase(object):
                 raise NoSuchDatabase(database)
         self.db = self._server[database]
         self.record_factory = record_factory or Record
+        self._changes_since = self.db.info()["update_seq"]
+        self._changes_last_used = 0  # Immediate run works.
+
 
     def _temporary_query(self, map_fun, reduce_fun=None, language='javascript',
             wrapper=None, **options):
@@ -139,41 +168,143 @@ class CouchDatabaseBase(object):
         except ResourceNotFound:
             return None
         data = {}
-        if 'deleted' in couch_record.get('application_annotations', {}).get(
-            'Ubuntu One', {}).get('private_application_annotations', {}):
+        if row_is_deleted(couch_record):
             return None
         data.update(couch_record)
         record = self.record_factory(data=data)
-        record.record_id = record_id
+
+        def make_getter(source_db, document_id, attachment_name, content_type):
+            """Closure storing the database for lower levels to use when needed.
+            """
+            def getter():
+                return source_db.get_attachment(document_id, attachment_name), \
+                        content_type
+            return getter
+        if "_attachments" in data:
+            for att_name, att_attributes in data["_attachments"].iteritems():
+                record.attach_by_reference(att_name,
+                        make_getter(self.db, record_id, att_name,
+                                att_attributes["content_type"]))
         return record
 
     def put_record(self, record):
         """Put a record in back end storage."""
-        record_id = record.record_id or record._data.get('_id', '')
-        record_data = record._data
-        if record_id:
-            self.db[record_id] = record_data
-        else:
-            record_id = self._add_record(record_data)
-        return record_id
+        if not record.record_id:
+            # Do not rely on couchdb to create an ID for us.
+            from uuid import uuid4
+            record.record_id = uuid4().hex
+        self.db[record.record_id] = record._data
 
-    def update_fields(self, record_id, fields):
+        for attachment_name in record.list_attachments():
+            data, content_type = record.attachment_data(attachment_name)
+            self.db.put_attachment(
+                record._data, data, attachment_name, content_type)
+
+        return record.record_id
+
+    def put_records_batch(self, batch):
+        """Put a batch of records in back end storage."""
+        # used to access fast the record
+        records_hash = {}
+        for record in batch:
+            if not record.record_id:
+                # Do not rely on couchdb to create an ID for us.
+                from uuid import uuid4
+                record.record_id = uuid4().hex
+            records_hash[record.record_id] = record
+        # although with a single record we need to test for the
+        # revisison, with a batch we do not, but we have to make sure
+        # that we did not get an error
+        batch_put_result = self.db.update([record._data for record in batch])
+        for current_tuple in batch_put_result:
+            success, docid, rev_or_exc = current_tuple
+            if success:
+                record = records_hash[docid]
+                # set the new rev
+                record._data["_rev"] = rev_or_exc
+                for attachment_name in record.list_attachments():
+                    data, content_type = record.attachment_data(attachment_name)
+                    self.db.put_attachment(
+                        {"_id":record.record_id, "_rev":record["_rev"]},
+                        data, attachment_name, content_type)
+        # all success record have the blobs added we return result of
+        # update
+        return batch_put_result
+
+    def update_fields(self, record_id, fields, cached_record=None):
         """Safely update a number of fields. 'fields' being a
-        dictionary with fieldname: value for only the fields we want
-        to change the value of.
-        """
-        while True:
-            record = self.db[record_id]
-            record.update(fields)
-            try:
-                self.db[record_id] = record
-            except ResourceConflict:
-                continue
-            break
+        dictionary with path_tuple: new_value for only the fields we
+        want to change the value of, where path_tuple is a tuple of
+        fieldnames indicating the path to the possibly nested field
+        we're interested in. old_record a the copy of the record we
+        most recently read from the database.
 
-    def _add_record(self, data):
-        """Add a new record to the storage backend."""
-        return self.db.create(data)
+        In the case the underlying document was changed, we try to
+        merge, but only if none of the old values have changed. (i.e.,
+        do not overwrite changes originating elsewhere.)
+
+        This is slightly hairy, so that other code won't have to be.
+        """
+        # Initially, the record in memory and in the db are the same
+        # as far as we know. (If they're not, we'll get a
+        # ResourceConflict later on, from which we can recover.)
+        if cached_record is None:
+            cached_record = self.db[record_id]
+        if isinstance(cached_record, Record):
+            cached_record = cached_record._data
+        record = copy.deepcopy(cached_record)
+        # Loop until either failure or success has been determined
+        while True:
+            modified = False
+            conflicts = {}
+            # loop through all the fields that need to be modified
+            for path, new_value in fields.items():
+                if not isinstance(path, tuple):
+                    path = (path,)
+                # Walk down in both copies of the record to the leaf
+                # node that represents the field, creating the path in
+                # the in memory record if necessary.
+                db_parent = record
+                cached_parent = cached_record
+                for field in path[:-1]:
+                    db_parent = db_parent.setdefault(field, {})
+                    cached_parent = cached_parent.get(field, {})
+                # Get the values of the fields in the two copies.
+                db_value = db_parent.get(path[-1])
+                cached_value = cached_parent.get(path[-1])
+                # If the value we intend to store is already in the
+                # database, we need do nothing, which is our favorite.
+                if db_value == new_value:
+                    continue
+                # If the value in the db is different than the value
+                # our copy holds, we have a conflict. We could bail
+                # here, but we can give better feedback if we gather
+                # all the conflicts, so we continue the for loop
+                if db_value != cached_value:
+                    conflicts[path] = (db_value, new_value)
+                    continue
+                # Otherwise, it is safe to update the field with the
+                # new value.
+                modified = True
+                db_parent[path[-1]] = new_value
+            # If we had conflicts, we can bail.
+            if conflicts:
+                raise FieldsConflict(conflicts)
+            # If we made changes to the document, we'll need to save
+            # it.
+            if modified:
+                try:
+                    self.db[record_id] = record
+                except ResourceConflict:
+                    # We got a conflict, meaning the record has
+                    # changed in the database since we last loaded it
+                    # into memory. Let's get a fresh copy and try
+                    # again.
+                    record = self.db[record_id]
+                    continue
+            # If we get here, nothing remains to be done, and we can
+            # take a well deserved break.
+            break
 
     def delete_record(self, record_id):
         """Delete record with given id"""
@@ -188,8 +319,7 @@ class CouchDatabaseBase(object):
         if record_id not in self.db:
             return False
         record = self.db[record_id]
-        return 'deleted' not in record.get('application_annotations', {}).get(
-            'Ubuntu One', {}).get('private_application_annotations', {})
+        return not row_is_deleted(record)
 
     def delete_view(self, view_name, design_doc=DEFAULT_DESIGN_DOCUMENT):
         """Remove a view, given its name.  Raises a KeyError on a unknown
@@ -234,13 +364,14 @@ class CouchDatabaseBase(object):
 
         return deleted_data
 
-    def execute_view(self, view_name, design_doc=DEFAULT_DESIGN_DOCUMENT):
+    def execute_view(self, view_name, design_doc=DEFAULT_DESIGN_DOCUMENT,
+            **params):
         """Execute view and return results."""
         if design_doc is None:
             design_doc = view_name
 
         view_id_fmt = "_design/%(design_doc)s/_view/%(view_name)s"
-        return self.db.view(view_id_fmt % locals())
+        return self.db.view(view_id_fmt % locals(), **params)
 
     def add_view(self, view_name, map_js, reduce_js,
             design_doc=DEFAULT_DESIGN_DOCUMENT):
@@ -332,3 +463,62 @@ class CouchDatabaseBase(object):
             return viewdata
         else:
             return viewdata[record_type]
+
+    def get_changes(self, niceness=10):
+        """Get a list of database changes.  This is the sister function of
+        report_changes that returns a list instead of calling a function for
+        each."""
+        l = list()
+        self.report_changes(lambda **k: l.append(k), niceness=niceness)
+        return l
+
+    def report_changes(self, function_to_send_changes_to, niceness=10):
+        """Check to see if there are any changes on this database since last
+        call (or since this object was instantiated), call a function for each,
+        and return the number of changes reported.
+
+        The callback function is called for every single change, with the
+        keyword parameters of the dictionary of values returned from couchdb.
+
+        >>> def f(seq=None, id=None, changes=None):
+        ...    pass
+
+        >>> db_foo.report_changes(f)
+        >>> time.sleep(30)
+        >>> db_foo.report_changes(f)
+
+        or
+
+        >>> # Make function that returns true, to signal never to remove.
+        >>> report_forever = lambda **kw: db_foo.report_changes(**kw) or True
+        >>> cb_id = glib.mainloop.idle_add(report_forever, f)
+
+        or
+
+        >>> task_id = twisted.task.looping_call(db_foo.report_changes, f)
+        """
+        now = time()
+        call_count = 0
+        if not niceness or now > self._changes_last_used + niceness:
+
+            # Can't use self._server.resource.get() directly because
+            # it encodes "/".
+            uri = couchdburi(
+                self._server.resource.uri, self.db.name, "_changes",
+                since=self._changes_since)
+            resp, data = self._server.resource.http.request(uri, "GET", "", {})
+            if resp["status"] != '200':
+                raise IOError(
+                    "HTTP response code %s.\n%s" % (resp["status"], data))
+            structure = json.loads(data)
+            for change in structure.get("results"):
+                # kw-args can't have unicode keys
+                change_encoded_keys = dict(
+                    (k.encode("utf8"), v) for k, v in change.iteritems())
+                function_to_send_changes_to(**change_encoded_keys)
+                # Not s.last_seq.  Exceptions!
+                self._changes_since = change["seq"]
+                call_count += 1
+            # If exception in cb, we never update governor.
+            self._changes_last_used = now
+        return call_count
