@@ -13,7 +13,9 @@ def is_active():
 
 # Required
 def oauth_data():
-    """OAuth information needed to replicate to a server."""
+    """OAuth information needed to replicate to a server.  This may be
+    called from a subthread, so be sure not to violate any execution 
+    idioms, like updating the GUI from a non-main thread."""
     return dict(consumer_key="", consumer_secret="", oauth_token="",
             oauth_token_secret="")
     # or to symbolize failure
