@@ -1,4 +1,3 @@
-#!/usr/bin/python
 # Copyright 2009 Canonical Ltd.
 #
 # This file is part of desktopcouch.
@@ -16,38 +15,27 @@
 # along with desktopcouch.  If not, see <http://www.gnu.org/licenses/>.
 #
 # Author: Chad Miller <chad.miller@canonical.com>
+#         Stuart Langridge <stuart.langridge@canonical.com>
+
 """
 Stop local CouchDB server.
 """
 
-import os
-import desktopcouch
-import time
-import signal
-import errno
+import subprocess, sys
+from desktopcouch import local_files
 
-def stop_couchdb(pid=None):
-    by_pid = pid
-    if pid is None:
-        pid = desktopcouch.find_pid(start_if_not_running=False)
-    while pid is not None:
-        try:
-            os.kill(pid, signal.SIGTERM)
-        except OSError, e:
-            if e.errno == errno.ESRCH:
-                break
-            raise
+def stop_couchdb(ctx=local_files.DEFAULT_CONTEXT):
+    local_exec = ctx.couch_exec_command + ["-k"]
+    try:
+        retcode = subprocess.call(local_exec, shell=False)
+        if retcode < 0:
+            print >> sys.stderr, "Child was terminated by signal", -retcode
+        elif retcode > 0:
+            print >> sys.stderr, "Child returned", retcode
+    except OSError, e:
+        print >> sys.stderr, "Execution failed: %s: %s" % (e, local_exec)
+        exit(1)
 
-        for retry in xrange(300):
-            try:
-                os.kill(pid, 0)  # test existence. sig-zero is special.
-            except OSError:
-                break
-            time.sleep(0.01)
-        if by_pid:
-            pid = None
-        else:
-            pid = desktopcouch.find_pid(start_if_not_running=False)
 
 if __name__ == "__main__":
     stop_couchdb()
